from __future__ import annotations

from collections.abc import Callable
import dataclasses
import functools
import os
import tomllib

from typing import Any, Self

import jinja2

import jinjarope

from jinjarope import envconfig, envglobals, envtests, loaders, utils


@functools.cache
def _load(txt: str) -> dict:
    """Some dumb caching since this might be invoked on each env instanciation."""
    return tomllib.loads(txt)


class JinjaFile(dict):
    """A file defining filters / tests."""

    def __init__(self, path: str | os.PathLike):
        """Instanciate the file.

        Arguments:
            path: Path to the jinja file
        """
        super().__init__()
        text = envglobals.load_file_cached(os.fspath(path))
        data = _load(text)
        self.update(data)

    @property
    def filters(self) -> list[JinjaItem]:
        """Return list of filters defined in the file."""
        return [
            JinjaItem(filter_name, **dct)
            for filter_name, dct in self.get("filters", {}).items()
            if all(envtests.is_installed(i) for i in dct.get("required_packages", []))
        ]

    @property
    def tests(self) -> list[JinjaItem]:
        """Return list of tests defined in the file."""
        return [
            JinjaItem(filter_name, **dct)
            for filter_name, dct in self.get("tests", {}).items()
            if all(envtests.is_installed(i) for i in dct.get("required_packages", []))
        ]

    @property
    def functions(self) -> list[JinjaItem]:
        """Return list of functions defined in the file."""
        return [
            JinjaItem(filter_name, **dct)
            for filter_name, dct in self.get("functions", {}).items()
            if all(envtests.is_installed(i) for i in dct.get("required_packages", []))
        ]

    @property
    def filters_dict(self) -> dict[str, Callable]:
        """Return a dictionary with all filters.

        Can directly get merged into env filters.
        """
        dct = {}
        for f in self.filters:
            dct[f.identifier] = f.filter_fn
            for alias in f.aliases:
                dct[alias] = f.filter_fn
        return dct

    @property
    def tests_dict(self) -> dict[str, Callable]:
        """Return a dictionary with all filters.

        Can directly get merged into env filters.
        """
        dct = {}
        for f in self.tests:
            dct[f.identifier] = f.filter_fn
            for alias in f.aliases:
                dct[alias] = f.filter_fn
        return dct

    @property
    def functions_dict(self) -> dict[str, Callable]:
        """Return a dictionary with all filters.

        Can directly get merged into env filters.
        """
        dct = {}
        for f in self.functions:
            dct[f.identifier] = f.filter_fn
            for alias in f.aliases:
                dct[alias] = f.filter_fn
        return dct

    @property
    def envconfig(self) -> envconfig.EnvConfig:
        """Return the config object defined in this Jinja file."""
        cfg = self.get("config", {})
        return envconfig.EnvConfig(**cfg)

    @property
    def loader(self) -> jinja2.BaseLoader | None:
        """Return a (composed Choice-) loader defined in this Jinja file."""
        return loaders.from_json(self.get("loaders", []))


@dataclasses.dataclass(frozen=True)
class JinjaItem:
    """An item representing a filter / test."""

    identifier: str
    fn: str
    group: str
    examples: dict[str, dict] = dataclasses.field(default_factory=dict)
    description: str | None = None
    aliases: list[str] = dataclasses.field(default_factory=list)
    required_packages: list[str] = dataclasses.field(default_factory=list)

    @property
    def filter_fn(self) -> Callable:
        """Return the callable to use as filter / test / function."""
        try:
            obj = utils.resolve(self.fn)
        except AttributeError:
            msg = f"Could not import jinja item {self.identifier!r} from {self.fn!r}"
            raise ImportError(msg) from AttributeError
        if not callable(obj):
            msg = "Filter needs correct, importable Path for callable"
            raise TypeError(msg)
        return obj

    @classmethod
    def for_function(cls, fn: Callable, group: str = "imported", **kwargs: Any) -> Self:
        """Alternative ctor to construct a JinjaItem based on a callable.

        Arguments:
            fn: Callable to get a JinjaItem for
            group: Group for metadata
            kwargs: Additional keyword arguments for JinjaItem ctor
        """
        return cls(fn.__name__, f"{fn.__module__}.{fn.__name__}", group=group, **kwargs)

    def apply(self, *args: Any, **kwargs: Any) -> Any:
        """Apply the filter function using given arguments and keywords.

        Arguments:
            args: The arguments for the call
            kwargs: They keyword arguments for the call
        """
        return self.filter_fn(*args, **kwargs)

    def resolve_example(self, example_name: str) -> str:
        """Render example with given name and return the result.

        Arguments:
            example_name: The example identifier
        """
        example = self.examples[example_name]
        env = jinjarope.Environment(loader=jinjarope.FileSystemLoader(""))
        return env.render_string(example["template"])


if __name__ == "__main__":
    file = JinjaFile("src/jinjarope/resources/filters.toml")
    print(file.filters[5].resolve_example("basic"))



from __future__ import annotations

from collections.abc import Sequence
import datetime
import functools
import logging
import os
import pathlib

from typing import Any

from jinjarope import utils


logger = logging.getLogger(__name__)


@functools.cache
def load_file_cached(path: str | os.PathLike) -> str:
    """Return the str-content of file at given path.

    Arguments:
        path: The path to get str content from
    """
    if "://" in str(path):
        return utils.fsspec_get(str(path))
    return pathlib.Path(path).read_text(encoding="utf-8")


_cache: dict[str, str] = {}


def get_output_from_call(
    call: str | Sequence[str],
    cwd: str | os.PathLike | None = None,
    use_cache: bool = False,
) -> str | None:
    """Execute a system call and return its output as a string.

    Arguments:
        call: The system call to make
        cwd: The working directory for the call
        use_cache: Whether to cache the output of calls
    """
    import subprocess

    if not call:
        return None
    if not isinstance(call, str):
        call = " ".join(call)
    key = pathlib.Path(cwd or ".").absolute().as_posix() + call
    if key in _cache and use_cache:
        return _cache[key]
    msg = f"Executing {call!r}..."
    logger.info(msg)
    try:
        pipe = subprocess.PIPE
        text = subprocess.run(call, stdout=pipe, text=True, shell=True, cwd=cwd).stdout
        _cache[key] = text
        return text  # noqa: TRY300
    except subprocess.CalledProcessError:
        logger.warning("Executing %s failed", call)
        return None


def add(text: str | None, prefix: str = "", suffix: str = "") -> str:
    """Add a pre- or suffix to a value if the value is true-ish.

    Arguments:
        text: The text to check
        prefix: Prefix to add if text is true-ish
        suffix: Suffix to add if text is true-ish
    """
    return f"{prefix}{text}{suffix}" if text else ""


def ternary(value: Any, true_val: Any, false_val: Any, none_val: Any = None):
    """Value ? true_val : false_val.

    Arguments:
        value: The value to check.
        true_val: The value to return if given value is true-ish
        false_val: The value to return if given value is false-ish
        none_val: Optional value to return if given value is None
    """
    if value is None and none_val is not None:
        return none_val
    return true_val if bool(value) else false_val


def match(obj, mapping: dict | None = None, **kwargs: Any) -> str:
    """A filter trying to imitate a python match-case statement.

    Arguments:
        obj: match object
        mapping: a mapping for the different cases. If key is type, an isinstance will
                 be performed. If key is a str, check for equality.
        kwargs: Same functionality as mapping, but provided as keyword arguments for
                convenience.

    Examples:
        ``` jinja
        {{ "a" | match(a="hit", b="miss")
        {{ MyClass() | match({MyClass: "hit", OtherClass: "miss"}) }}
    """
    # kwargs can only contain strs as keys, so we can perform simply getitem.
    if kwargs and obj in kwargs:
        return kwargs[obj]

    for k, v in (mapping or {}).items():
        match k:
            case type() if isinstance(obj, k):
                return v
            case _ if k == obj:
                return v
    return ""


def has_internet() -> bool:
    """Return true if machine is connected to internet.

    Checks connection with a HEAD request to the Google DNS server.
    """
    import http.client as httplib

    conn = httplib.HTTPSConnection("8.8.8.8", timeout=2)
    try:
        conn.request("HEAD", "/")
        return True  # noqa: TRY300
    except Exception:  # noqa: BLE001
        return False
    finally:
        conn.close()


def now(tz: datetime.tzinfo | None = None) -> datetime.datetime:
    """Get current Datetime.

    Arguments:
        tz: timezone for retuned datetime
    """
    return datetime.datetime.now(tz)


def utcnow() -> datetime.datetime:
    """Get UTC datetime."""
    return datetime.datetime.utcnow()


ENV_GLOBALS = {
    "range": range,
    "zip": zip,
    "set": set,
    "tuple": tuple,
    "list": list,
    "int": int,
    "str": str,
}



from __future__ import annotations

from collections.abc import Callable, Mapping
import os
import pathlib

from typing import Any, Literal

import jinja2

from jinjarope import iterfilters, loaders, serializefilters, utils


class NestedDictLoader(loaders.LoaderMixin, jinja2.BaseLoader):
    """A jinja loader for loading templates from nested dicts.

    This loader allows to access templates from nested dicts.
    Can be used to load templates defined with markup like TOML.

    Examples:
        ``` toml
        [example]
        template = "{{ something }}"
        ```
        ``` py
        content = tomllib.load(toml_file)
        loader = NestedDictLoader(content)
        env = Environment(loader=loader)
        env.get_template("example/template")
        ```
    """

    ID = "nested_dict"

    def __init__(self, mapping: Mapping):
        """Constructor.

        Arguments:
            mapping: A nested dict containing templates
        """
        super().__init__()
        self._data = mapping

    def __repr__(self):
        return utils.get_repr(self, mapping=self._data)

    def list_templates(self) -> list[str]:
        return list(iterfilters.flatten_dict(self._data).keys())

    def get_source(
        self,
        environment: jinja2.Environment,
        template: str,
    ) -> tuple[str, str, Callable[[], bool] | None]:
        data: Any = self._data
        try:
            for part in template.split("/"):
                data = data[part]
            assert isinstance(data, str)
        except (AssertionError, KeyError) as e:
            raise jinja2.TemplateNotFound(template) from e
        return data, None, lambda: True  # type: ignore[return-value]


class TemplateFileLoader(NestedDictLoader):
    """A jinja loader for loading templates from config files.

    This loader allows to access templates from config files.
    Config files often often resemble nested dicts when getting loaded / deserialized.

    The loader will load config file from given path and will make it accessible in the
    same way as the `NestedDictLoader`. (esp. TOML is well-suited for this)

    Config files can be loaded from any fsspec protocol URL.

    Examples:
        ``` py
        loader = TemplateFileLoader("http://path_to_toml_file.toml")
        env = Environment(loader=loader)
        env.get_template("example/template")
        ```
        ``` py
        loader = TemplateFileLoader("path/to/file.json")
        env = Environment(loader=loader)
        env.get_template("example/template")
        ```
    """

    ID = "template_file"

    def __init__(
        self,
        path: str | os.PathLike,
        fmt: Literal["toml", "json", "ini", "yaml"] | None = None,
    ):
        """Constructor.

        Arguments:
            path: Path / fsspec protocol URL to the file
            fmt: Config file format. If None, try to auto-infer from file extension
        """
        self.path = os.fspath(path)
        text = utils.fsspec_get(self.path)
        file_fmt = fmt if fmt else pathlib.Path(self.path).suffix.lstrip(".")
        assert file_fmt in ["json", "toml", "yaml", "ini"]
        mapping = serializefilters.deserialize(text, fmt=file_fmt)  # type: ignore[arg-type]
        super().__init__(mapping=mapping)
        self._data = mapping

    def __repr__(self):
        return utils.get_repr(self, path=pathlib.Path(self.path).as_posix())


if __name__ == "__main__":
    from jinjarope import Environment

    env = Environment()
    env.loader = NestedDictLoader({"a": {"b": "c"}})
    text = env.render_template("a/b")
    print(text)



from __future__ import annotations

from collections.abc import Mapping
import json
import re

from typing import Any, Literal
from xml.etree import ElementTree as Et


def wrap_in_elem(
    text: str | None,
    tag: str,
    add_linebreaks: bool = False,
    **kwargs: Any,
) -> str:
    """Wrap given text in an HTML/XML tag (with attributes).

    If text is empty, just return an empty string.

    Arguments:
        text: Text to wrap
        tag: Tag to wrap text in
        add_linebreaks: Adds a linebreak before and after the text
        kwargs: additional key-value pairs to be inserted as attributes for tag.
                Key strings will have "_" stripped from the end to allow using keywords.
    """
    if not text:
        return ""
    attrs = [f'{k.rstrip("_")}="{v}"' for k, v in kwargs.items()]
    attr_str = (" " + " ".join(attrs)) if attrs else ""
    nl = "\n" if add_linebreaks else ""
    return f"<{tag}{attr_str}>{nl}{text}{nl}</{tag}>"


def html_link(text: str | None = None, link: str | None = None, **kwargs: Any) -> str:
    """Create a html link.

    If link is empty string or None, just the text will get returned.

    Arguments:
        text: Text to show for the link
        link: Target url
        kwargs: additional key-value pairs to be inserted as attributes.
                Key strings will have "_" stripped from the end to allow using keywords.
    """
    if not link:
        return text or ""
    attrs = [f'{k.rstrip("_")}="{v}"' for k, v in kwargs.items()]
    attr_str = (" " + " ".join(attrs)) if attrs else ""
    return f"<a href={link!r}{attr_str}>{text or link}</a>"


def format_js_map(mapping: dict | str, indent: int = 4) -> str:
    """Return JS map str for given dictionary.

    Arguments:
        mapping: Dictionary to dump
        indent: The amount of indentation for the key-value pairs
    """
    dct = json.loads(mapping) if isinstance(mapping, str) else mapping
    rows = []
    indent_str = " " * indent
    for k, v in dct.items():
        match v:
            case bool():
                rows.append(f"{indent_str}{k}: {str(v).lower()},")
            case dict():
                rows.append(f"{indent_str}{k}: {format_js_map(v)},")
            case None:
                rows.append(f"{indent_str}{k}: null,")
            case _:
                rows.append(f"{indent_str}{k}: {v!r},")
    row_str = "\n" + "\n".join(rows) + "\n"
    return f"{{{row_str}}}"


def svg_to_data_uri(svg: str) -> str:
    """Wrap svg as data URL.

    Arguments:
        svg: The svg to wrap into a data URL
    """
    if not isinstance(svg, str):
        msg = "Invalid type: %r"
        raise TypeError(msg, type(svg))
    return f"url('data:image/svg+xml;charset=utf-8,{ svg }')"


def clean_svg(text: str) -> str:
    """Strip off unwanted stuff from svg text which might be added by external libs.

    Removes xml headers and doctype declarations.

    Arguments:
        text: The text to cleanup / filter
    """
    text = re.sub(r"<\?xml version.*\?>\s*", "", text, flags=re.DOTALL)
    text = re.sub(r"<!DOCTYPE svg.*?>", "", text, flags=re.DOTALL)
    return text.strip()


def format_css_rule(dct: Mapping) -> str:
    """Format a nested dictionary as CSS rule.

    Mapping must be of shape {".a": {"b": "c"}}

    Arguments:
        dct: The mapping to convert to CSS text
    """
    data: dict[str, list] = {}

    def _parse(obj, selector: str = ""):
        for key, value in obj.items():
            if hasattr(value, "items"):
                rule = selector + " " + key
                data[rule] = []
                _parse(value, rule)

            else:
                prop = data[selector]
                prop.append(f"\t{key}: {value};\n")

    _parse(dct)
    string = ""
    for key, value in sorted(data.items()):
        if data[key]:
            string += key[1:] + " {\n" + "".join(value) + "}\n\n"
    return string


def format_xml(
    str_or_elem: str | Et.Element,
    indent: str | int = "  ",
    level: int = 0,
    method: Literal["xml", "html", "text", "c14n"] = "html",
    short_empty_elements: bool = True,
    add_declaration: bool = False,
) -> str:
    """(Pretty)print given XML.

    Arguments:
        str_or_elem: XML to prettyprint
        indent: Amount of spaces to use for indentation
        level: Initial indentation level
        method: Output method
        short_empty_elements: Whether empty elements should get printed in short form
                              (applies when mode is "xml")
        add_declaration: whether a XML declaration should be printed
                         (applies when mode is "xml")
    """
    if isinstance(str_or_elem, str):
        str_or_elem = Et.fromstring(str_or_elem)
    space = indent if isinstance(indent, str) else indent * " "
    Et.indent(str_or_elem, space=space, level=level)
    return Et.tostring(
        str_or_elem,
        encoding="unicode",
        method=method,
        xml_declaration=add_declaration,
        short_empty_elements=short_empty_elements,
    )


if __name__ == "__main__":
    print(format_js_map({"key": {"nested_key": "nested_value"}}))



from __future__ import annotations

import pathlib

import jinja2
import jinjarope
import pytest


def test_environment_init():
    env = jinjarope.Environment(
        undefined="strict",
        trim_blocks=True,
        cache_size=-1,
        auto_reload=False,
    )

    assert isinstance(env, jinja2.Environment)
    assert env.undefined == jinja2.StrictUndefined
    assert env.trim_blocks is True
    assert env.auto_reload is False


def test_environment_contains():
    env = jinjarope.Environment(
        undefined="strict",
        trim_blocks=True,
        cache_size=-1,
        auto_reload=False,
    )
    env.loader = jinja2.DictLoader({"home.html": "Home", "about.html": "About"})

    assert "home.html" in env
    assert "about.html" in env
    assert "nonexistent.html" not in env


def test_getitem():
    env = jinjarope.Environment()
    env.loader = jinja2.DictLoader({"home.html": "Home", "about.html": "About"})

    assert env["home.html"].render() == "Home"
    assert env["about.html"].render() == "About"
    with pytest.raises(jinja2.exceptions.TemplateNotFound):
        env["nonexistent.html"]


def test_compile():
    env = jinjarope.Environment()
    source = "Hello, {{ name }}!"
    compiled = env.compile(source)
    assert isinstance(compiled, type(compile("", "", "exec")))


def test_inherit_from():
    env1 = jinjarope.Environment()
    env1.globals["global_var"] = "Global Variable"
    env2 = jinjarope.Environment()
    env2.inherit_from(env1)
    assert "global_var" in env2.globals
    assert env2.globals["global_var"] == "Global Variable"


def test_add_template():
    env = jinjarope.Environment()
    env.loader = jinja2.DictLoader({})
    env.add_template("tests/testresources/testfile.jinja")
    assert "tests/testresources/testfile.jinja" in env
    assert env["tests/testresources/testfile.jinja"].render() == "CONTENT!"


def test_add_template_path():
    env = jinjarope.Environment()
    env.add_template_path("templates")
    assert "templates" in env._extra_paths
    assert isinstance(env.loader, jinja2.loaders.FileSystemLoader)


def test__add_loader():
    env = jinjarope.Environment()
    env._add_loader("templates")
    assert isinstance(env.loader, jinja2.loaders.FileSystemLoader)
    env._add_loader({"home.html": "Home"})
    assert isinstance(env.loader, jinja2.loaders.ChoiceLoader)
    assert isinstance(env.loader.loaders[0], jinja2.loaders.DictLoader)
    assert isinstance(env.loader.loaders[1], jinja2.loaders.FileSystemLoader)


def test_render_condition():
    env = jinjarope.Environment()
    assert env.render_condition("{{ 1 == 1 }}") is True
    assert env.render_condition("{{ 1 == 2 }}") is False
    assert env.render_condition("{{ var == 'value' }}", {"var": "value"}) is True
    assert env.render_condition("{{ var == 'value' }}", {"var": "other"}) is False


def test_render_string():
    env = jinjarope.Environment()
    result = env.render_string("Hello, {{ name }}!", {"name": "World"})
    assert result == "Hello, World!"


def test_render_file(tmp_path: pathlib.Path):
    env = jinjarope.Environment()
    file = tmp_path / "test.txt"
    file.write_text("Hello, {{ name }}!")
    result = env.render_file(file, {"name": "World"})
    assert result == "Hello, World!"


def test_render_template():
    env = jinjarope.Environment()
    env.loader = jinja2.DictLoader({"home.html": "Hello, {{ name }}!"})
    result = env.render_template("home.html", {"name": "World"})
    assert result == "Hello, World!"


def test_render_template_block():
    env = jinjarope.Environment()
    dct = {"home.html": "{% block greeting %}Hello, {{ name }}!{% endblock %}"}
    env.loader = jinja2.DictLoader(dct)
    result = env.render_template("home.html", {"name": "World"}, block_name="greeting")
    assert result == "Hello, World!"


def test_render_template_block_not_found():
    env = jinjarope.Environment()
    dct = {"home.html": "{% block greeting %}Hello, {{ name }}!{% endblock %}"}
    env.loader = jinja2.DictLoader(dct)
    with pytest.raises(jinjarope.BlockNotFoundError):
        env.render_template("home.html", {"name": "World"}, block_name="farewell")


def test_with_globals():
    env = jinjarope.Environment()
    assert "name" not in env.globals
    with env.with_globals(name="World"):
        assert env.globals["name"] == "World"
        print(list(env.globals.keys()))
    assert "name" not in env.globals


def test_setup_loader():
    env = jinjarope.Environment()
    env.setup_loader(dir_paths=["testresources"])
    assert isinstance(env.loader, jinja2.loaders.ChoiceLoader)


def test_evaluate():
    env = jinjarope.Environment()
    result = env.evaluate("1 + 1")
    assert result == 2  # noqa: PLR2004


def test_get_config():
    env = jinjarope.Environment(trim_blocks=True)
    config = env.get_config()
    assert config.trim_blocks is True



from __future__ import annotations

from jinjarope import manual
import mknodes as mk


def test_building_the_docs():
    build = manual.Build()
    nav = mk.MkNav()
    build.on_root(nav)
    assert nav.children
    for node in nav.descendants:
        if isinstance(node, mk.MkPage):
            str(node)



from __future__ import annotations

import ast
import contextlib
import io
import logging
import os
import pathlib
import time

from types import CodeType
from typing import Any, Literal, overload
import weakref

import jinja2

from jinja2.exceptions import TemplateSyntaxError
import jinja2.nodes

import jinjarope

from jinjarope import (
    envconfig,
    envglobals,
    jinjafile,
    loaders,
    undefined as undefined_,
    utils,
)


logger = logging.getLogger(__name__)


class Environment(jinja2.Environment):
    """An enhanced Jinja environment."""

    def __init__(
        self,
        *,
        undefined: undefined_.UndefinedStr | type[jinja2.Undefined] = "strict",
        trim_blocks: bool = True,
        cache_size: int = -1,
        auto_reload: bool = False,
        loader: (
            jinja2.BaseLoader | list[jinja2.BaseLoader] | dict | list[dict] | None
        ) = None,
        **kwargs: Any,
    ):
        """Constructor.

        Arguments:
            undefined: Handling of "Undefined" errors
            trim_blocks: Whitespace handling. Changes jinja default to `True`.
            cache_size: Amount of templates to cache.
                        Changes jinja default to not clean cache.
            auto_reload: Whether to check templates for changes on loading
            loader: Loader to use (Also accepts a JSON representation of loaders)
            kwargs: Keyword arguments passed to parent
        """
        self.cache_code = True
        if isinstance(undefined, str):
            undefined = undefined_.UNDEFINED_BEHAVIOR[undefined]
        kwargs = dict(
            undefined=undefined,
            trim_blocks=trim_blocks,
            auto_reload=auto_reload,
            cache_size=cache_size,
            loader=loaders.from_json(loader),
            **kwargs,
        )
        self._extra_files: set[str] = set()
        self._extra_paths: set[str] = set()
        super().__init__(**kwargs)

        # Update namespaces
        folder = pathlib.Path(__file__).parent / "resources"
        self.load_jinja_file(folder / "filters.toml")
        self.load_jinja_file(folder / "tests.toml")
        self.load_jinja_file(folder / "functions.toml")
        self.globals.update(envglobals.ENV_GLOBALS)
        for fn in utils._entry_points("jinjarope.environment").values():
            fn(self)
        self.filters["render_template"] = self.render_template
        self.filters["render_string"] = self.render_string
        self.filters["render_file"] = self.render_file
        self.filters["evaluate"] = self.evaluate
        self.globals["filters"] = self.filters
        self.globals["tests"] = self.tests
        self.tests["template"] = lambda template_name: template_name in self
        self.template_cache: weakref.WeakValueDictionary[
            str | jinja2.nodes.Template,
            CodeType | str | None,
        ] = weakref.WeakValueDictionary()
        self.add_extension("jinja2.ext.loopcontrols")
        self.add_extension("jinja2.ext.do")

    def __contains__(self, template: str | os.PathLike) -> bool:
        """Check whether given template path exists.

        Arguments:
            template: The template path to check
        """
        return pathlib.Path(template).as_posix() in self.list_templates()

    def __getitem__(self, val: str) -> jinja2.Template:
        """Return a template by path.

        val: The template path
        """
        return self.get_template(val)

    def set_undefined(self, value: undefined_.UndefinedStr | type[jinja2.Undefined]):
        """Set the undefined behaviour for the environment.

        Arguments:
            value: The new undefined behaviour
        """
        new = undefined_.UNDEFINED_BEHAVIOR[value] if isinstance(value, str) else value
        self.undefined = new

    def load_jinja_file(
        self,
        path: str | os.PathLike,
        scope_prefix: str = "",
        load_filters: bool = True,
        load_tests: bool = True,
        load_functions: bool = True,
        load_config: bool = True,
        load_loader: bool = True,
    ):
        """Load the content of a JinjaFile and add it to the environment.

        Arguments:
            path: The path to the JinjaFile
            scope_prefix: Optional prefix to add to all tests / filters / functions
            load_filters: Whether to load filters from the JinjaFile
            load_tests: Whether to load tests from the JinjaFile
            load_functions: Whether to load functions from the JinjaFile
            load_config: Whether to load the environment config from the JinjaFile
            load_loader: Whether to load the Loader from the JinjaFile
        """
        file = jinjafile.JinjaFile(path)
        if load_filters:
            dct = {f"{scope_prefix}{k}": v for k, v in file.filters_dict.items()}
            self.filters.update(dct)
        if load_tests:
            dct = {f"{scope_prefix}{k}": v for k, v in file.tests_dict.items()}
            self.tests.update(dct)
        if load_functions:
            dct = {f"{scope_prefix}{k}": v for k, v in file.functions_dict.items()}
            self.globals.update(dct)
        if load_config:
            self.__dict__.update(file.envconfig.as_dict())
        if load_loader and (loader := file.loader):
            self._add_loader(loader)

    @overload
    def compile(  # type: ignore[misc]  # noqa: A003
        self,
        source: str | jinja2.nodes.Template,
        name: str | None = None,
        filename: str | None = None,
        raw: Literal[False] = False,
        defer_init: bool = False,
    ) -> CodeType: ...

    @overload
    def compile(  # noqa: A003
        self,
        source: str | jinja2.nodes.Template,
        name: str | None = None,
        filename: str | None = None,
        raw: Literal[True] = ...,
        defer_init: bool = False,
    ) -> str: ...

    def compile(  # noqa: A003
        self,
        source: str | jinja2.nodes.Template,
        name: str | None = None,
        filename: str | None = None,
        raw: bool = False,
        defer_init: bool = False,
    ) -> CodeType | str:
        """Compile the template."""
        if (
            not self.cache_code
            or name is not None
            or filename is not None
            or raw is not False
            or defer_init is not False
        ):
            # If there are any non-default keywords args, we do
            # not cache.
            return super().compile(  # type: ignore[no-any-return,call-overload]
                source,
                name,
                filename,
                raw,
                defer_init,
            )

        if (cached := self.template_cache.get(source)) is None:
            cached = self.template_cache[source] = super().compile(source)

        return cached

    def inherit_from(self, env: jinja2.Environment):
        """Inherit complete configuration from another environment.

        Arguments:
            env: The environment to inherit settings from
        """
        self.__dict__.update(env.__dict__)
        self.linked_to = env
        self.overlayed = True

    def add_template(self, file: str | os.PathLike):
        """Add a new template during runtime.

        Will create a new DictLoader and inject it into the existing loaders.

        Useful since render_string/render_file does not allow to use a parent template.
        Using this, render_template can be used.

        Arguments:
            file: File to add as a template
        """
        # we keep track of already added extra files to not add things multiple times.
        file = str(file)
        if file in self._extra_files:
            return
        self._extra_files.add(file)
        content = envglobals.load_file_cached(file)
        new_loader = loaders.DictLoader({file: content})
        self._add_loader(new_loader)

    def add_template_path(self, *path: str | os.PathLike):
        """Add a new template path during runtime.

        Will append a new FileSystemLoader by wrapping it and the the current loader into
        either an already-existing or a new Choiceloader.

        Arguments:
            path: Template serch path(s) to add
        """
        for p in path:
            if p in self._extra_paths:
                return
            self._extra_paths.add(str(p))
            new_loader = loaders.FileSystemLoader(p)
            self._add_loader(new_loader)

    def _add_loader(self, new_loader: jinja2.BaseLoader | dict | str | os.PathLike):
        match new_loader:
            case dict():
                new_loader = loaders.DictLoader(new_loader)
            case str() | os.PathLike():
                new_loader = loaders.FileSystemLoader(new_loader)
        match self.loader:
            case jinja2.ChoiceLoader():
                self.loader.loaders = [new_loader, *self.loader.loaders]
            case None:
                self.loader = new_loader
            case _:
                self.loader = loaders.ChoiceLoader(loaders=[new_loader, self.loader])

    def render_condition(
        self,
        string: str,
        variables: dict | None = None,
        **kwargs: Any,
    ) -> bool:
        """Render a template condition.

        Returns True for true-ish return values from a render_string call.

        Arguments:
            string: String to evaluate for True-ishness
            variables: Extra variables for the rendering
            kwargs: Further extra variables for rendering
        """
        result = self.render_string(string=string, variables=variables, **kwargs)
        return result not in ["None", "False", ""]

    def render_string(
        self,
        string: str,
        variables: dict | None = None,
        **kwargs: Any,
    ) -> str:
        """Render a template string.

        Arguments:
            string: String to render
            variables: Extra variables for the rendering
            kwargs: Further extra variables for rendering
        """
        variables = (variables or {}) | kwargs
        cls = self.template_class
        try:
            template = cls.from_code(self, self.compile(string), self.globals, None)
        except TemplateSyntaxError as e:
            msg = f"Error when evaluating \n{string}\n (extra globals: {variables})"
            raise SyntaxError(msg) from e
        return template.render(**variables)

    def render_file(
        self,
        file: str | os.PathLike,
        variables: dict | None = None,
        **kwargs: Any,
    ) -> str:
        """Helper to directly render a template from filesystem.

        Note: The file we pull in gets cached. That should be fine for our case though.

        Arguments:
            file: Template file to load
            variables: Extra variables for the rendering
            kwargs: Further extra variables for rendering
        """
        content = envglobals.load_file_cached(str(file))
        return self.render_string(content, variables, **kwargs)

    def render_template(
        self,
        template_name: str,
        variables: dict[str, Any] | None = None,
        block_name: str | None = None,
        parent_template: str | None = None,
        **kwargs: Any,
    ) -> str:
        """Render a loaded template (or a block of a template).

        Arguments:
            template_name: Template name
            variables: Extra variables for rendering
            block_name: Render specific block from the template
            parent_template: The name of the parent template importing this template
            kwargs: Further extra variables for rendering
        """
        variables = (variables or {}) | kwargs
        template = self.get_template(template_name, parent=parent_template)
        if not block_name:
            return template.render(**variables)
        try:
            block_render_func = template.blocks[block_name]
        except KeyError:
            raise BlockNotFoundError(block_name, template_name) from KeyError

        ctx = template.new_context(variables)
        return self.concat(block_render_func(ctx))  # type: ignore
        # except Exception:
        #     self.handle_exception()

    @contextlib.contextmanager
    def with_globals(self, **kwargs: Any):
        """Context manager to temporarily set globals for the environment.

        Arguments:
            kwargs: Globals to set
        """
        temp = self.globals.copy()
        self.globals.update(kwargs)
        yield
        self.globals = temp

    def setup_loader(
        self,
        dir_paths: list[str] | None = None,
        module_paths: list[str] | None = None,
        static: dict[str, str] | None = None,
        fsspec_paths: bool = True,
    ):
        self.loader = jinjarope.get_loader(
            dir_paths=dir_paths,
            module_paths=module_paths,
            static=static,
            fsspec_paths=fsspec_paths,
        )

    def evaluate(
        self,
        code: str,
        context: dict[str, Any] | None = None,
    ) -> str:
        """Evaluate python code and return the caught stdout + return value of last line.

        Arguments:
            code: The code to execute
            context: Globals for the execution evironment
        """
        now = time.time()
        logger.debug("Evaluating code:\n%s", code)
        tree = ast.parse(code)
        eval_expr = ast.Expression(tree.body[-1].value)  # type: ignore
        # exec_expr = ast.Module(tree.body[:-1])  # type: ignore
        exec_expr = ast.parse("")
        exec_expr.body = tree.body[:-1]
        compiled = compile(exec_expr, "file", "exec")
        buffer = io.StringIO()
        with contextlib.redirect_stdout(buffer):
            exec(compiled, self.globals)
            val = eval(compile(eval_expr, "file", "eval"), self.globals)
        logger.debug("Code evaluation took %s seconds.", time.time() - now)
        # result = mk.MkContainer([buffer.getvalue(), val])
        return val or ""

    def get_config(self) -> envconfig.EnvConfig:
        """All environment settings as a dict (not included: undefined and loaders)."""
        return envconfig.EnvConfig(
            block_start_string=self.block_start_string,
            block_end_string=self.block_end_string,
            variable_start_string=self.variable_start_string,
            variable_end_string=self.variable_end_string,
            comment_start_string=self.comment_start_string,
            comment_end_string=self.comment_end_string,
            line_statement_prefix=self.line_statement_prefix,
            line_comment_prefix=self.line_comment_prefix,
            trim_blocks=self.trim_blocks,
            lstrip_blocks=self.lstrip_blocks,
            newline_sequence=self.newline_sequence,
            keep_trailing_newline=self.keep_trailing_newline,
        )


class BlockNotFoundError(Exception):
    """Exception for not-found template blocks."""

    def __init__(
        self,
        block_name: str,
        template_name: str,
        message: str | None = None,
    ):
        self.block_name = block_name
        self.template_name = template_name
        super().__init__(
            message
            or f"Block {self.block_name!r} not found in template {self.template_name!r}",
        )


if __name__ == "__main__":
    env = Environment()
    txt = """{% filter indent %}
    test
    {% endfilter %}
    """
    print(env.render_string(txt))



from __future__ import annotations

import jinjarope


def test_prefix_loader():
    loader = "test" / jinjarope.DictLoader({"a": "b"})
    env = jinjarope.Environment(loader=loader)
    assert env.render_template("test/a") == "b"
    assert "test/a" in loader
    loader1 = jinjarope.PrefixLoader({"prefix": jinjarope.DictLoader({})})
    loader2 = jinjarope.PrefixLoader({"prefix": jinjarope.DictLoader({})})
    assert loader1 == loader2
    assert len({loader1, loader2}) == 1
    assert repr(loader1)


def test_function_loader():
    def load_fn(path: str) -> str:
        return "abc"

    loader1 = jinjarope.FunctionLoader(load_fn)
    loader2 = jinjarope.FunctionLoader(load_fn)
    assert loader1 == loader2
    assert len({loader1, loader2}) == 1
    assert repr(loader1)


def test_packageloader():
    loader1 = jinjarope.PackageLoader("jinjarope")
    loader2 = jinjarope.PackageLoader(jinjarope)
    assert loader1 == loader2
    assert len({loader1, loader2}) == 1
    assert repr(loader1)


def test_filesystemloader():
    loader1 = jinjarope.FileSystemLoader("")
    loader2 = jinjarope.FileSystemLoader("")
    assert loader1 == loader2
    assert len({loader1, loader2}) == 1
    assert repr(loader1)



from __future__ import annotations

import typer as t

import jinjarope


CFG_HELP = "JinjaFiles to load."
OUT_HELP = "Output path for resolved template. If not given, print to stdout."
IN_HELP = "Path to a (remote) template file."
UNDEF_HELP = "Set undefined behavior. Overrides JinjaFiles settings."
TRIM_HELP = "Trim blocks. Overrides JinjaFiles settings."

CFG_CMDS = ["-j", "--jinja-file"]
OUT_CMDS = ["-o", "--output-file"]
IN_CMDS = ["-i", "--input-file"]
UNDEF_CMDS = ["-u", "--undefined"]
TRIM_CMDS = ["-t", "--trim-blocks"]


cli = t.Typer(
    name="JinjaRope",
    help=(
        " 🚀🚀🚀 JinjaRope CLI interface. Render templates from the terminal! 🚀🚀🚀\n\n"
        "Check out https://phil65.github.io/jinjarope/ !"
    ),
    no_args_is_help=True,
)


@cli.command()
def render(
    template_path: str = t.Option(
        None,
        *IN_CMDS,
        help=IN_HELP,
        show_default=False,
    ),
    cfg_files: list[str] = t.Option(  # noqa: B008
        None,
        *CFG_CMDS,
        help=CFG_HELP,
        show_default=False,
    ),
    output: str = t.Option(
        None,
        *OUT_CMDS,
        help=OUT_HELP,
        show_default=False,
    ),
    undefined: str = t.Option(
        "strict",
        *UNDEF_CMDS,
        help=UNDEF_HELP,
        show_default=False,
    ),
    trim_blocks: bool = t.Option(
        None,
        *TRIM_CMDS,
        help=TRIM_HELP,
        show_default=False,
    ),
):
    env = jinjarope.Environment()
    for path in cfg_files:
        env.load_jinja_file(path)
    if undefined:
        env.set_undefined(undefined)  # type: ignore[arg-type]
    if trim_blocks is not None:
        env.trim_blocks = trim_blocks
    text = env.render_file(template_path)
    if output is None:
        print(text)
    else:
        import fsspec

        with fsspec.open(output, mode="w", encoding="utf-8") as f:
            f.write(text)


if __name__ == "__main__":
    cli(
        [
            "-i",
            "github://phil65:mknodes@main/docs/icons.jinja",
            "-j",
            "src/jinjarope/resources/tests.toml",
            "-j",
            "src/jinjarope/resources/filters.toml",
            "-o",
            "out.test",
        ],
    )



from __future__ import annotations

from collections.abc import Callable
import pathlib

import jinja2

from jinjarope import loaders as loaders_, utils


class RewriteLoader(loaders_.LoaderMixin, jinja2.BaseLoader):
    """A loader which modifies templates based on a callable.

    Can get chained like a PrefixLoader.
    The path passed to the callable can be used to check whether given template
    should be modified.
    """

    ID = "rewrite"

    def __init__(self, loader: jinja2.BaseLoader, rewrite_fn: Callable[[str, str], str]):
        """Instanciate the RewriteLoader.

        Arguments:
            loader: The loader to rewrite / modify the templates from
            rewrite_fn: Callable to modify templates.
                        It gets called with two arguments (path and template text)
                        and should return a (possibly modified) template text
        """
        self.loader = loader
        self.rewrite_fn = rewrite_fn

    def __repr__(self):
        return utils.get_repr(self, self.loader, self.rewrite_fn)

    def __eq__(self, other):
        return (
            type(self) == type(other)
            and self.loader == other.loader
            and self.rewrite_fn == other.rewrite_fn
        )

    def __hash__(self):
        return hash(self.loader) + hash(self.rewrite_fn)

    def get_source(
        self,
        environment: jinja2.Environment,
        template: str,
    ) -> tuple[str, str, Callable[[], bool] | None]:
        src: str | None
        src, filename, uptodate = self.loader.get_source(environment, template)
        old_src = src
        assert filename is not None
        path = pathlib.Path(filename).as_posix()
        src = self.rewrite_fn(path, src)
        return src or old_src, filename, uptodate


if __name__ == "__main__":
    import jinjarope

    loader = jinjarope.FsSpecProtocolPathLoader()
    rewrite_loader = RewriteLoader(loader, lambda path, x: x.replace("a", "XX"))
    env = jinjarope.Environment()
    env.loader = rewrite_loader
    result = env.render_template("github://phil65:mknodes@main/README.md")
    print(result)
